@@ -11,15 +11,23 @@ Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)
 This source code is licensed under the MIT license found in the LICENSE
 file in the root directory of this source tree.
 """
+from __future__ import annotations
+
 import copy
-from abc import ABCMeta, abstractmethod
+from abc import ABCMeta
+from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union, cast
 
 import fs
 import numpy as np
+from fs.base import FS
 
+from sentinelhub import SHConfig
+
+from .constants import FeatureType
 from .eodata import EOPatch
 from .eotask import EOTask
 from .utils.fs import get_filesystem, pickle_fs, unpickle_fs
+from .utils.parsing import FeatureSpec, FeaturesSpecification
 
 
 class CopyTask(EOTask):
@@ -28,39 +36,36 @@ class CopyTask(EOTask):
     It copies feature type dictionaries but not the data itself.
     """
 
-    def __init__(self, features=...):
+    def __init__(self, features: FeaturesSpecification = ...):
         """
         :param features: A collection of features or feature types that will be copied into a new EOPatch.
-        :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         """
         self.features = features
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         return eopatch.copy(features=self.features)
 
 
 class DeepCopyTask(CopyTask):
     """Makes a deep copy of the given EOPatch."""
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         return eopatch.copy(features=self.features, deep=True)
 
 
-class IOTask(EOTask, metaclass=ABCMeta):
+class IOTask(EOTask, metaclass=ABCMeta):  # noqa B024
     """An abstract Input/Output task that can handle a path and a filesystem object."""
 
-    def __init__(self, path, filesystem=None, create=False, config=None):
+    def __init__(
+        self, path: str, filesystem: Optional[FS] = None, create: bool = False, config: Optional[SHConfig] = None
+    ):
         """
         :param path: root path where all EOPatches are saved
-        :type path: str
         :param filesystem: An existing filesystem object. If not given it will be initialized according to the EOPatch
             path.
-        :type filesystem: fs.base.FS or None
         :param create: If the filesystem path doesn't exist this flag indicates to either create it or raise an error
-        :type create: bool
         :param config: A configuration object with AWS credentials. By default, is set to None and in this case the
             default configuration will be taken.
-        :type config: SHConfig or None
         """
         self.path = path
         self.filesystem_path = "/" if filesystem is None else self.path
@@ -70,7 +75,7 @@ class IOTask(EOTask, metaclass=ABCMeta):
         self.config = config
 
     @property
-    def filesystem(self):
+    def filesystem(self) -> FS:
         """A filesystem property that unpickles an existing filesystem definition or creates a new one."""
         if self._pickled_filesystem is None:
             filesystem = get_filesystem(self.path, create=self._create_path, config=self.config)
@@ -79,24 +84,17 @@ class IOTask(EOTask, metaclass=ABCMeta):
 
         return unpickle_fs(self._pickled_filesystem)
 
-    @abstractmethod
-    def execute(self, *eopatches, **kwargs):
-        """Implement execute function"""
-
 
 class SaveTask(IOTask):
     """Saves the given EOPatch to a filesystem."""
 
-    def __init__(self, path, filesystem=None, config=None, **kwargs):
+    def __init__(self, path: str, filesystem: Optional[FS] = None, config: Optional[SHConfig] = None, **kwargs: Any):
         """
         :param path: root path where all EOPatches are saved
-        :type path: str
         :param filesystem: An existing filesystem object. If not given it will be initialized according to the EOPatch
             path.
-        :type filesystem: fs.base.FS or None
         :param features: A collection of features types specifying features of which type will be saved. By default,
             all features will be saved.
-        :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param overwrite_permission: A level of permission for overwriting an existing EOPatch
         :type overwrite_permission: OverwritePermission or int
         :param compress_level: A level of data compression and can be specified with an integer from 0 (no compression)
@@ -104,20 +102,16 @@ class SaveTask(IOTask):
         :type compress_level: int
         :param config: A configuration object with AWS credentials. By default, is set to None and in this case the
             default configuration will be taken.
-        :type config: SHConfig or None
         """
         self.kwargs = kwargs
         super().__init__(path, filesystem=filesystem, create=True, config=config)
 
-    def execute(self, eopatch, *, eopatch_folder=""):
+    def execute(self, eopatch: EOPatch, *, eopatch_folder: Optional[str] = "") -> EOPatch:
         """Saves the EOPatch to disk: `folder/eopatch_folder`.
 
         :param eopatch: EOPatch which will be saved
-        :type eopatch: EOPatch
         :param eopatch_folder: Name of EOPatch folder containing data. If `None` is given it won't save anything.
-        :type eopatch_folder: str or None
         :return: The same EOPatch
-        :rtype: EOPatch
         """
         if eopatch_folder is None:
             return eopatch
@@ -130,34 +124,28 @@ class SaveTask(IOTask):
 class LoadTask(IOTask):
     """Loads an EOPatch from a filesystem."""
 
-    def __init__(self, path, filesystem=None, config=None, **kwargs):
+    def __init__(self, path: str, filesystem: Optional[FS] = None, config: Optional[SHConfig] = None, **kwargs: Any):
         """
         :param path: root directory where all EOPatches are saved
-        :type path: str
         :param filesystem: An existing filesystem object. If not given it will be initialized according to the EOPatch
             path. If you intend to run this task in multiprocessing mode you shouldn't specify this parameter.
-        :type filesystem: fs.base.FS or None
         :param features: A collection of features to be loaded. By default, all features will be loaded.
-        :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param lazy_loading: If `True` features will be lazy loaded. Default is `False`
         :type lazy_loading: bool
         :param config: A configuration object with AWS credentials. By default, is set to None and in this case the
             default configuration will be taken.
-        :type config: SHConfig or None
         """
         self.kwargs = kwargs
         super().__init__(path, filesystem=filesystem, create=False, config=config)
 
-    def execute(self, eopatch=None, *, eopatch_folder=""):
+    def execute(self, eopatch: Optional[EOPatch] = None, *, eopatch_folder: Optional[str] = "") -> EOPatch:
         """Loads the EOPatch from disk: `folder/eopatch_folder`.
 
         :param eopatch: Optional input EOPatch. If given the loaded features are merged onto it, otherwise a new EOPatch
             is created.
         :param eopatch_folder: Name of EOPatch folder containing data. If `None` is given it will return an empty
             or modified `EOPatch` (depending on the task input).
-        :type eopatch_folder: str or None
         :return: EOPatch loaded from disk
-        :rtype: EOPatch
         """
         if eopatch_folder is None:
             return eopatch or EOPatch()
@@ -172,22 +160,18 @@ class LoadTask(IOTask):
 class AddFeatureTask(EOTask):
     """Adds a feature to the given EOPatch."""
 
-    def __init__(self, feature):
+    def __init__(self, feature: FeatureSpec):
         """
         :param feature: Feature to be added
-        :type feature: (FeatureType, feature_name) or FeatureType
         """
         self.feature_type, self.feature_name = self.parse_feature(feature)
 
-    def execute(self, eopatch, data):
+    def execute(self, eopatch: EOPatch, data: object) -> EOPatch:
         """Returns the EOPatch with added features.
 
         :param eopatch: input EOPatch
-        :type eopatch: EOPatch
         :param data: data to be added to the feature
-        :type data: object
         :return: input EOPatch with the specified feature
-        :rtype: EOPatch
         """
         if self.feature_name is None:
             eopatch[self.feature_type] = data
@@ -200,23 +184,20 @@ class AddFeatureTask(EOTask):
 class RemoveFeatureTask(EOTask):
     """Removes one or multiple features from the given EOPatch."""
 
-    def __init__(self, features):
+    def __init__(self, features: FeaturesSpecification):
         """
         :param features: A collection of features to be removed.
-        :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         """
         self.feature_parser = self.get_feature_parser(features)
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """Returns the EOPatch with removed features.
 
         :param eopatch: input EOPatch
-        :type eopatch: EOPatch
         :return: input EOPatch without the specified feature
-        :rtype: EOPatch
         """
         for feature_type, feature_name in self.feature_parser.get_features(eopatch):
-            if feature_name is ...:
+            if feature_name is None:
                 eopatch.reset_feature_type(feature_type)
             else:
                 del eopatch[feature_type][feature_name]
@@ -227,20 +208,17 @@ class RemoveFeatureTask(EOTask):
 class RenameFeatureTask(EOTask):
     """Renames one or multiple features from the given EOPatch."""
 
-    def __init__(self, features):
+    def __init__(self, features: FeaturesSpecification):
         """
         :param features: A collection of features to be renamed.
-        :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         """
         self.feature_parser = self.get_feature_parser(features)
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """Returns the EOPatch with renamed features.
 
         :param eopatch: input EOPatch
-        :type eopatch: EOPatch
         :return: input EOPatch with the renamed features
-        :rtype: EOPatch
         """
         for feature_type, feature_name, new_feature_name in self.feature_parser.get_renamed_features(eopatch):
             eopatch[feature_type][new_feature_name] = eopatch[feature_type][feature_name]
@@ -252,23 +230,19 @@ class RenameFeatureTask(EOTask):
 class DuplicateFeatureTask(EOTask):
     """Duplicates one or multiple features in an EOPatch."""
 
-    def __init__(self, features, deep_copy=False):
+    def __init__(self, features: FeaturesSpecification, deep_copy: bool = False):
         """
         :param features: A collection of features to be copied.
-        :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param deep_copy: Make a deep copy of feature's data if set to true, else just assign it.
-        :type deep_copy: bool
         """
         self.feature_parser = self.get_feature_parser(features)
         self.deep = deep_copy
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """Returns the EOPatch with copied features.
 
         :param eopatch: Input EOPatch
-        :type eopatch: EOPatch
         :return: Input EOPatch with the duplicated features.
-        :rtype: EOPatch
         :raises ValueError: Raises an exception when trying to duplicate a feature with an
             already existing feature name.
         """
@@ -298,20 +272,24 @@ class InitializeFeatureTask(EOTask):
         InitializeFeature((FeatureType.MASK, 'mask1'), shape=(FeatureType.DATA, 'data1'), init_value=1)
     """
 
-    def __init__(self, features, shape, init_value=0, dtype=np.uint8):
+    def __init__(
+        self,
+        features: FeaturesSpecification,
+        shape: Union[Tuple[int, ...], FeatureSpec],
+        init_value: int = 0,
+        dtype: Union[np.dtype, type] = np.uint8,
+    ):
         """
         :param features: A collection of features to initialize.
-        :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param shape: A shape object (t, n, m, d) or a feature from which to read the shape.
-        :type shape: A tuple or an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param init_value: A value with which to initialize the array of the new feature.
-        :type init_value: int
         :param dtype: Type of array values.
-        :type dtype: NumPy dtype
         :raises ValueError: Raises an exception when passing the wrong shape argument.
         """
 
         self.features = self.parse_features(features)
+        self.shape_feature: Optional[Tuple[FeatureType, Optional[str]]]
+        self.shape: Union[None, Tuple[int, int, int], Tuple[int, int, int, int]]
 
         try:
             self.shape_feature = self.parse_feature(shape)
@@ -321,23 +299,21 @@ class InitializeFeatureTask(EOTask):
         if self.shape_feature:
             self.shape = None
         elif isinstance(shape, tuple) and len(shape) in (3, 4) and all(isinstance(x, int) for x in shape):
-            self.shape = shape
+            self.shape = cast(Union[Tuple[int, int, int], Tuple[int, int, int, int]], shape)
         else:
             raise ValueError("shape argument is not a shape tuple or a feature containing one.")
 
         self.init_value = init_value
         self.dtype = dtype
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """
         :param eopatch: Input EOPatch.
-        :type eopatch: EOPatch
         :return: Input EOPatch with the initialized additional features.
-        :rtype: EOPatch
         """
         shape = eopatch[self.shape_feature].shape if self.shape_feature else self.shape
 
-        add_features = set(self.features) - set(eopatch.get_feature_list())
+        add_features = set(self.features) - set(self.parse_features(eopatch.get_feature_list()))
 
         for feature in add_features:
             eopatch[feature] = np.ones(shape, dtype=self.dtype) * self.init_value
@@ -348,24 +324,19 @@ class InitializeFeatureTask(EOTask):
 class MoveFeatureTask(EOTask):
     """Task to copy/deepcopy fields from one EOPatch to another."""
 
-    def __init__(self, features, deep_copy=False):
+    def __init__(self, features: FeaturesSpecification, deep_copy: bool = False):
         """
         :param features: A collection of features to be moved.
-        :type features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param deep_copy: Make a deep copy of feature's data if set to true, else just assign it.
-        :type deep_copy: bool
         """
         self.feature_parser = self.get_feature_parser(features)
         self.deep = deep_copy
 
-    def execute(self, src_eopatch, dst_eopatch):
+    def execute(self, src_eopatch: EOPatch, dst_eopatch: EOPatch) -> EOPatch:
         """
         :param src_eopatch: Source EOPatch from which to take features.
-        :type src_eopatch: EOPatch
         :param dst_eopatch: Destination EOPatch to which to move/copy features.
-        :type dst_eopatch: EOPatch
         :return: dst_eopatch with the additional features from src_eopatch.
-        :rtype: EOPatch
         """
 
         for feature in self.feature_parser.get_features(src_eopatch):
@@ -416,34 +387,39 @@ class MapFeatureTask(EOTask):
         result = maximum(patch)
     """
 
-    def __init__(self, input_features, output_features, map_function=None, **kwargs):
+    def __init__(
+        self,
+        input_features: FeaturesSpecification,
+        output_features: FeaturesSpecification,
+        map_function: Optional[Callable] = None,
+        **kwargs: Any,
+    ):
         """
         :param input_features: A collection of the input features to be mapped.
-        :type input_features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param output_features: A collection of the output features to which to assign the output data.
-        :type output_features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param map_function: A function or lambda to be applied to the input data.
         :raises ValueError: Raises an exception when passing feature collections with different lengths.
         :param kwargs: kwargs to be passed to the map function.
         """
         self.input_features = self.parse_features(input_features)
-        self.output_feature = self.parse_features(output_features)
+        self.output_features = self.parse_features(output_features)
         self.kwargs = kwargs
 
-        if len(self.input_features) != len(self.output_feature):
+        if len(self.input_features) != len(self.output_features):
             raise ValueError("The number of input and output features must match.")
 
-        self.function = map_function if map_function else self.map_method
+        if map_function:  # mypy 0.981 has issues with inlined conditional and functions
+            self.function: Callable = map_function
+        else:
+            self.function = self.map_method
 
-    def execute(self, eopatch):
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """
         :param eopatch: Source EOPatch from which to read the data of input features.
-        :type eopatch: EOPatch
         :return: An eopatch with the additional mapped features.
-        :rtype: EOPatch
         """
-        for input_features, output_feature in zip(self.input_features, self.output_feature):
-            eopatch[output_feature] = self.function(eopatch[input_features], **self.kwargs)
+        for input_feature, output_feature in zip(self.input_features, self.output_features):
+            eopatch[output_feature] = self.function(eopatch[input_feature], **self.kwargs)
 
         return eopatch
 
@@ -493,26 +469,32 @@ class ZipFeatureTask(EOTask):
             result = maximum(patch)
     """
 
-    def __init__(self, input_features, output_feature, zip_function=None, **kwargs):
+    def __init__(
+        self,
+        input_features: FeaturesSpecification,
+        output_feature: FeaturesSpecification,
+        zip_function: Optional[Callable] = None,
+        **kwargs: Any,
+    ):
         """
         :param input_features: A collection of the input features to be mapped.
-        :type input_features: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param output_feature: An output feature object to which to assign the data.
-        :type output_feature: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param zip_function: A function or lambda to be applied to the input data.
         :param kwargs: kwargs to be passed to the zip function.
         """
         self.input_features = self.parse_features(input_features)
         self.output_feature = self.parse_feature(output_feature)
-        self.function = zip_function if zip_function else self.zip_method
         self.kwargs = kwargs
 
-    def execute(self, eopatch):
+        if zip_function:  # mypy 0.981 has issues with inlined conditional and functions
+            self.function: Callable = zip_function
+        else:
+            self.function = self.zip_method
+
+    def execute(self, eopatch: EOPatch) -> EOPatch:
         """
         :param eopatch: Source EOPatch from which to read the data of input features.
-        :type eopatch: EOPatch
         :return: An eopatch with the additional zipped features.
-        :rtype: EOPatch
         """
         data = [eopatch[feature] for feature in self.input_features]
 
@@ -528,7 +510,7 @@ class ZipFeatureTask(EOTask):
 class MergeFeatureTask(ZipFeatureTask):
     """Merges multiple features together by concatenating their data along the last axis."""
 
-    def zip_method(self, *f, dtype=None):
+    def zip_method(self, *f: np.ndarray, dtype: Union[None, np.dtype, type] = None) -> np.ndarray:
         """Concatenates the data of features along the last axis."""
         return np.concatenate(f, axis=-1, dtype=dtype)  # pylint: disable=unexpected-keyword-arg
 
@@ -536,14 +518,11 @@ class MergeFeatureTask(ZipFeatureTask):
 class ExtractBandsTask(MapFeatureTask):
     """Moves a subset of bands from one feature to a new one."""
 
-    def __init__(self, input_feature, output_feature, bands):
+    def __init__(self, input_feature: FeaturesSpecification, output_feature: FeaturesSpecification, bands: List[int]):
         """
         :param input_feature: A source feature from which to take the subset of bands.
-        :type input_feature: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param output_feature: An output feature to which to write the bands.
-        :type output_feature: an object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param bands: A list of bands to be moved.
-        :type bands: list
         """
         super().__init__(input_feature, output_feature)
         self.bands = bands
@@ -555,15 +534,38 @@ class ExtractBandsTask(MapFeatureTask):
         return feature[..., self.bands]
 
 
+class ExplodeBandsTask(EOTask):
+    """Explode a subset of bands from one feature to multiple new features."""
+
+    def __init__(
+        self,
+        input_feature: Tuple[FeatureType, str],
+        output_mapping: Dict[Tuple[FeatureType, str], Union[int, Iterable[int]]],
+    ):
+        """
+        :param input_feature: A source feature from which to take the subset of bands.
+        :param output_mapping: A mapping of output features into the band indices used to explode the input feature.
+        """
+        self.input_feature = input_feature
+        self.output_mapping = output_mapping
+
+    def execute(self, eopatch: EOPatch) -> EOPatch:
+        for output_feature, bands in self.output_mapping.items():
+            new_bands = list(bands) if isinstance(bands, Iterable) else [bands]
+            eopatch = ExtractBandsTask(
+                input_feature=self.input_feature, output_feature=output_feature, bands=new_bands
+            ).execute(eopatch)
+        return eopatch
+
+
 class CreateEOPatchTask(EOTask):
     """Creates an EOPatch."""
 
-    def execute(self, **kwargs):
+    def execute(self, **kwargs: Any) -> EOPatch:
         """Returns a newly created EOPatch with the given kwargs.
 
         :param kwargs: Any valid kwargs accepted by :class:`EOPatch.__init__<eolearn.core.eodata.EOPatch>`
         :return: A new eopatch.
-        :rtype: EOPatch
         """
         return EOPatch(**kwargs)
 
@@ -574,18 +576,16 @@ class MergeEOPatchesTask(EOTask):
     Check :func:`EOPatch.merge<eolearn.core.eodata.EOPatch.merge>` for more information about the merging process.
     """
 
-    def __init__(self, **merge_kwargs):
+    def __init__(self, **merge_kwargs: Any):
         """
         :param merge_kwargs: Keyword arguments defined for `EOPatch.merge` method.
         """
         self.merge_kwargs = merge_kwargs
 
-    def execute(self, *eopatches):
+    def execute(self, *eopatches: EOPatch) -> EOPatch:
         """
         :param eopatches: EOPatches to be merged
-        :type eopatches: EOPatch
         :return: A new EOPatch with merged content
-        :rtype: EOPatch
         """
         if not eopatches:
             raise ValueError("At least one EOPatch should be given")
