@@ -14,14 +14,16 @@ file in the root directory of this source tree.
 from __future__ import annotations
 
 import copy
-import datetime
+import datetime as dt
 import logging
-from typing import TYPE_CHECKING, List, Optional, Tuple, Union
+from abc import ABCMeta, abstractmethod
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple, Type, TypeVar, Union, cast, overload
 
 import attr
 import dateutil.parser
 import geopandas as gpd
 import numpy as np
+from fs.base import FS
 
 from sentinelhub import CRS, BBox
 
@@ -30,7 +32,11 @@ from .eodata_io import FeatureIO, load_eopatch, save_eopatch
 from .eodata_merge import merge_eopatches
 from .utils.common import deep_eq, is_discrete_type
 from .utils.fs import get_filesystem
-from .utils.parsing import parse_features
+from .utils.parsing import FeatureSpec, FeaturesSpecification, parse_features
+from .utils.types import EllipsisType, Literal
+
+_T = TypeVar("_T")
+_Self = TypeVar("_Self")
 
 LOGGER = logging.getLogger(__name__)
 
@@ -44,6 +50,170 @@ if TYPE_CHECKING:
         pass
 
 
+class _FeatureDict(Dict[str, Union[_T, FeatureIO[_T]]], metaclass=ABCMeta):
+    """A dictionary structure that holds features of certain feature type.
+
+    It checks that features have a correct and dimension. It also supports lazy loading by accepting a function as a
+    feature value, which is then called when the feature is accessed.
+    """
+
+    FORBIDDEN_CHARS = {".", "/", "\\", "|", ";", ":", "\n", "\t"}
+
+    def __init__(self, feature_dict: Dict[str, Union[_T, FeatureIO[_T]]], feature_type: FeatureType):
+        """
+        :param feature_dict: A dictionary of feature names and values
+        :param feature_type: Type of features
+        """
+        super().__init__()
+
+        self.feature_type = feature_type
+
+        for feature_name, value in feature_dict.items():
+            self[feature_name] = value
+
+    @classmethod
+    def empty_factory(cls: Type[_Self], feature_type: FeatureType) -> Callable[[], _Self]:
+        """Returns a factory function for creating empty feature dictionaries with an appropriate feature type."""
+
+        def factory() -> _Self:
+            return cls(feature_dict={}, feature_type=feature_type)  # type: ignore[call-arg]
+
+        return factory
+
+    def __setitem__(self, feature_name: str, value: Union[_T, FeatureIO[_T]]) -> None:
+        """Before setting value to the dictionary it checks that value is of correct type and dimension and tries to
+        transform value in correct form.
+        """
+        if not isinstance(value, FeatureIO):
+            value = self._parse_feature_value(value, feature_name)
+        self._check_feature_name(feature_name)
+        super().__setitem__(feature_name, value)
+
+    def _check_feature_name(self, feature_name: str) -> None:
+        """Ensures that feature names are strings and do not contain forbidden characters."""
+        if not isinstance(feature_name, str):
+            raise ValueError(f"Feature name must be a string but an object of type {type(feature_name)} was given.")
+
+        for char in feature_name:
+            if char in self.FORBIDDEN_CHARS:
+                raise ValueError(
+                    f"The name of feature ({self.feature_type}, {feature_name}) contains an illegal character '{char}'."
+                )
+
+        if feature_name == "":
+            raise ValueError("Feature name cannot be an empty string.")
+
+    @overload
+    def __getitem__(self, feature_name: str, load: Literal[True] = ...) -> _T:
+        ...
+
+    @overload
+    def __getitem__(self, feature_name: str, load: Literal[False] = ...) -> Union[_T, FeatureIO[_T]]:
+        ...
+
+    def __getitem__(self, feature_name: str, load: bool = True) -> Union[_T, FeatureIO[_T]]:
+        """Implements lazy loading."""
+        value = super().__getitem__(feature_name)
+
+        if isinstance(value, FeatureIO) and load:
+            value = value.load()
+            self[feature_name] = value
+
+        return value
+
+    def __eq__(self, other: object) -> bool:
+        """Compares its content against a content of another feature type dictionary."""
+        return deep_eq(self, other)
+
+    def __ne__(self, other: object) -> bool:
+        """Compares its content against a content of another feature type dictionary."""
+        return not self.__eq__(other)
+
+    def get_dict(self) -> Dict[str, _T]:
+        """Returns a Python dictionary of features and value."""
+        return dict(self)
+
+    @abstractmethod
+    def _parse_feature_value(self, value: object, feature_name: str) -> _T:
+        """Checks if value fits the feature type. If not it tries to fix it or raise an error.
+
+        :raises: ValueError
+        """
+
+
+class _FeatureDictNumpy(_FeatureDict[np.ndarray]):
+    """_FeatureDict object specialized for Numpy arrays."""
+
+    def __init__(self, feature_dict: Dict[str, Union[np.ndarray, FeatureIO[np.ndarray]]], feature_type: FeatureType):
+        ndim = feature_type.ndim()
+        if ndim is None:
+            raise ValueError(f"Feature type {feature_type} does not represent a Numpy based feature.")
+        self.ndim = ndim
+        super().__init__(feature_dict, feature_type)
+
+    def _parse_feature_value(self, value: object, feature_name: str) -> np.ndarray:
+        if not isinstance(value, np.ndarray):
+            raise ValueError(f"{self.feature_type} feature has to be a numpy array.")
+        if not hasattr(self, "ndim"):  # Because of serialization/deserialization during multiprocessing
+            return value
+        if value.ndim != self.ndim:
+            raise ValueError(
+                f"Numpy array of {self.feature_type} feature has to have {self.ndim} "
+                f"dimension{'s' if self.ndim > 1 else ''} but feature {feature_name} has {value.ndim}."
+            )
+
+        if self.feature_type.is_discrete() and not is_discrete_type(value.dtype):
+            raise ValueError(
+                f"{self.feature_type} is a discrete feature type therefore dtype of data array "
+                f"has to be either integer or boolean type but feature {feature_name} has dtype {value.dtype.type}."
+            )
+
+        return value
+
+
+class _FeatureDictGeoDf(_FeatureDict[gpd.GeoDataFrame]):
+    """_FeatureDict object specialized for GeoDataFrames."""
+
+    def __init__(self, feature_dict: Dict[str, gpd.GeoDataFrame], feature_type: FeatureType):
+        if not feature_type.is_vector:
+            raise ValueError(f"Feature type {feature_type} does not represent a vector feature.")
+        super().__init__(feature_dict, feature_type)
+
+    def _parse_feature_value(self, value: object, feature_name: str) -> gpd.GeoDataFrame:
+        if isinstance(value, gpd.GeoSeries):
+            value = gpd.GeoDataFrame(geometry=value, crs=value.crs)
+
+        if isinstance(value, gpd.GeoDataFrame):
+            if self.feature_type is FeatureType.VECTOR and FeatureType.TIMESTAMP.value.upper() not in value:
+                raise ValueError(
+                    f"{self.feature_type} feature has to contain a column 'TIMESTAMP' with timestamps but "
+                    f"feature {feature_name} doesn't not have it."
+                )
+
+            return value
+
+        raise ValueError(
+            f"{self.feature_type} feature works with data of type {gpd.GeoDataFrame.__name__} but feature "
+            f"{feature_name} has data of type {type(value)}."
+        )
+
+
+class _FeatureDictJson(_FeatureDict[Any]):
+    """_FeatureDict object specialized for meta-info."""
+
+    def _parse_feature_value(self, value: object, _: str) -> Any:
+        return value
+
+
+def _create_feature_dict(feature_type: FeatureType, value: Dict[str, Any]) -> _FeatureDict:
+    """Creates the correct FeatureDict, corresponding to the FeatureType."""
+    if feature_type.is_vector():
+        return _FeatureDictGeoDf(value, feature_type)
+    if feature_type is FeatureType.META_INFO:
+        return _FeatureDictJson(value, feature_type)
+    return _FeatureDictNumpy(value, feature_type)
+
+
 @attr.s(repr=False, eq=False, kw_only=True)
 class EOPatch:
     """The basic data object for multi-temporal remotely sensed data, such as satellite imagery and its derivatives.
@@ -65,27 +235,27 @@ class EOPatch:
     arrays in other attributes.
     """
 
-    data = attr.ib(factory=dict)
-    mask = attr.ib(factory=dict)
-    scalar = attr.ib(factory=dict)
-    label = attr.ib(factory=dict)
-    vector = attr.ib(factory=dict)
-    data_timeless = attr.ib(factory=dict)
-    mask_timeless = attr.ib(factory=dict)
-    scalar_timeless = attr.ib(factory=dict)
-    label_timeless = attr.ib(factory=dict)
-    vector_timeless = attr.ib(factory=dict)
-    meta_info = attr.ib(factory=dict)
-    bbox = attr.ib(default=None)
-    timestamp = attr.ib(factory=list)
+    data: _FeatureDictNumpy = attr.ib(factory=_FeatureDictNumpy.empty_factory(FeatureType.DATA))
+    mask: _FeatureDictNumpy = attr.ib(factory=_FeatureDictNumpy.empty_factory(FeatureType.MASK))
+    scalar: _FeatureDictNumpy = attr.ib(factory=_FeatureDictNumpy.empty_factory(FeatureType.SCALAR))
+    label: _FeatureDictNumpy = attr.ib(factory=_FeatureDictNumpy.empty_factory(FeatureType.LABEL))
+    vector: _FeatureDictGeoDf = attr.ib(factory=_FeatureDictGeoDf.empty_factory(FeatureType.VECTOR))
+    data_timeless: _FeatureDictNumpy = attr.ib(factory=_FeatureDictNumpy.empty_factory(FeatureType.DATA_TIMELESS))
+    mask_timeless: _FeatureDictNumpy = attr.ib(factory=_FeatureDictNumpy.empty_factory(FeatureType.MASK_TIMELESS))
+    scalar_timeless: _FeatureDictNumpy = attr.ib(factory=_FeatureDictNumpy.empty_factory(FeatureType.SCALAR_TIMELESS))
+    label_timeless: _FeatureDictNumpy = attr.ib(factory=_FeatureDictNumpy.empty_factory(FeatureType.LABEL_TIMELESS))
+    vector_timeless: _FeatureDictGeoDf = attr.ib(factory=_FeatureDictGeoDf.empty_factory(FeatureType.VECTOR_TIMELESS))
+    meta_info: _FeatureDictJson = attr.ib(factory=_FeatureDictJson.empty_factory(FeatureType.META_INFO))
+    bbox: Optional[BBox] = attr.ib(default=None)
+    timestamp: List[dt.datetime] = attr.ib(factory=list)
 
-    def __setattr__(self, key, value, feature_name=None):
+    def __setattr__(self, key: str, value: object, feature_name: Union[str, None, EllipsisType] = None) -> None:
         """Raises TypeError if feature type attributes are not of correct type.
 
         In case they are a dictionary they are cast to _FeatureDict class.
         """
         if feature_name not in (None, Ellipsis) and FeatureType.has_value(key):
-            self[key][feature_name] = value
+            self.__getattribute__(key)[feature_name] = value
             return
 
         if FeatureType.has_value(key) and not isinstance(value, FeatureIO):
@@ -95,14 +265,16 @@ class EOPatch:
         super().__setattr__(key, value)
 
     @staticmethod
-    def _parse_feature_type_value(feature_type, value):
+    def _parse_feature_type_value(
+        feature_type: FeatureType, value: object
+    ) -> Union[_FeatureDict, BBox, List[dt.date], None]:
         """Checks or parses value which will be assigned to a feature type attribute of `EOPatch`. If the value
         cannot be parsed correctly it raises an error.
 
         :raises: TypeError, ValueError
         """
         if feature_type.has_dict() and isinstance(value, dict):
-            return value if isinstance(value, _FeatureDict) else _FeatureDict(value, feature_type)
+            return value if isinstance(value, _FeatureDict) else _create_feature_dict(feature_type, value)
 
         if feature_type is FeatureType.BBOX:
             if value is None or isinstance(value, BBox):
@@ -110,19 +282,17 @@ class EOPatch:
             if isinstance(value, (tuple, list)) and len(value) == 5:
                 return BBox(value[:4], crs=value[4])
 
-        if feature_type is FeatureType.TIMESTAMP:
-            if isinstance(value, (tuple, list)):
-                return [
-                    timestamp if isinstance(timestamp, datetime.date) else dateutil.parser.parse(timestamp)
-                    for timestamp in value
-                ]
+        if feature_type is FeatureType.TIMESTAMP and isinstance(value, (tuple, list)):
+            return [
+                timestamp if isinstance(timestamp, dt.date) else dateutil.parser.parse(timestamp) for timestamp in value
+            ]
 
         raise TypeError(
             f"Attribute {feature_type} requires value of type {feature_type.type()} - "
             f"failed to parse given value {value}"
         )
 
-    def __getattribute__(self, key, load=True, feature_name=None):
+    def __getattribute__(self, key: str, load: bool = True, feature_name: Union[str, None, EllipsisType] = None) -> Any:
         """Handles lazy loading and can even provide a single feature from _FeatureDict."""
         value = super().__getattribute__(key)
 
@@ -132,33 +302,48 @@ class EOPatch:
             value = getattr(self, key)
 
         if feature_name not in (None, Ellipsis) and isinstance(value, _FeatureDict):
+            feature_name = cast(str, feature_name)  # the above check deals with ... and None
             return value[feature_name]
 
         return value
 
-    def __getitem__(self, feature_type):
+    @overload
+    def __getitem__(
+        self, feature_type: Union[Literal[FeatureType.BBOX], Tuple[Literal[FeatureType.BBOX], Any]]
+    ) -> BBox:
+        ...
+
+    @overload
+    def __getitem__(
+        self, feature_type: Union[Literal[FeatureType.TIMESTAMP], Tuple[Literal[FeatureType.TIMESTAMP], Any]]
+    ) -> List[dt.datetime]:
+        ...
+
+    @overload
+    def __getitem__(self, feature_type: Union[FeatureType, Tuple[FeatureType, Union[str, None, EllipsisType]]]) -> Any:
+        ...
+
+    def __getitem__(self, feature_type: Union[FeatureType, Tuple[FeatureType, Union[str, None, EllipsisType]]]) -> Any:
         """Provides features of requested feature type. It can also accept a tuple of (feature_type, feature_name).
 
         :param feature_type: Type of EOPatch feature
-        :type feature_type: FeatureType or str or (FeatureType, str)
-        :return: Dictionary of features
         """
         feature_name = None
         if isinstance(feature_type, tuple):
             self._check_tuple_key(feature_type)
             feature_type, feature_name = feature_type
 
-        return self.__getattribute__(FeatureType(feature_type).value, feature_name=feature_name)
+        ftype = FeatureType(feature_type).value
+        return self.__getattribute__(ftype, feature_name=feature_name)  # type: ignore[call-arg]
 
-    def __setitem__(self, feature_type, value):
+    def __setitem__(
+        self, feature_type: Union[FeatureType, Tuple[FeatureType, Union[str, None, EllipsisType]]], value: Any
+    ) -> None:
         """Sets a new dictionary / list to the given FeatureType. As a key it can also accept a tuple of
         (feature_type, feature_name).
 
         :param feature_type: Type of EOPatch feature
-        :type feature_type: FeatureType or str or (FeatureType, str)
         :param value: New dictionary or list
-        :type value: dict or list
-        :return: Dictionary of features
         """
         feature_name = None
         if isinstance(feature_type, tuple):
@@ -167,33 +352,29 @@ class EOPatch:
 
         return self.__setattr__(FeatureType(feature_type).value, value, feature_name=feature_name)
 
-    def __delitem__(self, feature):
+    def __delitem__(self, feature: Tuple[FeatureType, str]) -> None:
         """Deletes the selected feature.
 
         :param feature: EOPatch feature
-        :type feature: (FeatureType, str)
         """
         self._check_tuple_key(feature)
         feature_type, feature_name = feature
         del self[feature_type][feature_name]
 
     @staticmethod
-    def _check_tuple_key(key):
+    def _check_tuple_key(key: tuple) -> None:
         """A helper function that checks a tuple, which should hold (feature_type, feature_name)."""
         if len(key) != 2:
             raise ValueError(f"Given element should be a tuple of (feature_type, feature_name), but {key} found.")
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
         """True if FeatureType attributes, bbox, and timestamps of both EOPatches are equal by value."""
-        if not isinstance(self, type(other)):
+        if not isinstance(other, type(self)):
             return False
 
-        for feature_type in FeatureType:
-            if not deep_eq(self[feature_type], other[feature_type]):
-                return False
-        return True
+        return all(deep_eq(self[feature_type], other[feature_type]) for feature_type in FeatureType)
 
-    def __contains__(self, feature: Union[FeatureType, Tuple[FeatureType, str]]):
+    def __contains__(self, feature: object) -> bool:
         if isinstance(feature, FeatureType):
             return bool(self[feature])
         if isinstance(feature, tuple) and len(feature) == 2:
@@ -206,11 +387,11 @@ class EOPatch:
             "`(feature_type, feature_name)` tuples."
         )
 
-    def __add__(self, other):
+    def __add__(self, other: EOPatch) -> EOPatch:
         """Merges two EOPatches into a new EOPatch."""
         return self.merge(other)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         feature_repr_list = []
         for feature_type in FeatureType:
             content = self[feature_type]
@@ -233,12 +414,11 @@ class EOPatch:
         return f"{self.__class__.__name__}({feature_repr})"
 
     @staticmethod
-    def _repr_value(value):
+    def _repr_value(value: object) -> str:
         """Creates a representation string for different types of data.
 
         :param value: data in any type
         :return: representation string
-        :rtype: str
         """
         if isinstance(value, np.ndarray):
             return f"{EOPatch._repr_value_class(value)}(shape={value.shape}, dtype={value.dtype})"
@@ -267,16 +447,15 @@ class EOPatch:
         return repr(value)
 
     @staticmethod
-    def _repr_value_class(value):
+    def _repr_value_class(value: object) -> str:
         """A representation of a class of a given value"""
         cls = value.__class__
         return ".".join([cls.__module__.split(".")[0], cls.__name__])
 
-    def __copy__(self, features=...):
+    def __copy__(self, features: FeaturesSpecification = ...) -> EOPatch:
         """Returns a new EOPatch with shallow copies of given features.
 
         :param features: A collection of features or feature types that will be copied into new EOPatch.
-        :type features: object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         """
         if not features:  # For some reason deepcopy and copy pass {} by default
             features = ...
@@ -289,13 +468,11 @@ class EOPatch:
                 new_eopatch[feature_type] = copy.copy(self[feature_type])
         return new_eopatch
 
-    def __deepcopy__(self, memo=None, features=...):
+    def __deepcopy__(self, memo: Optional[dict] = None, features: FeaturesSpecification = ...) -> EOPatch:
         """Returns a new EOPatch with deep copies of given features.
 
         :param memo: built-in parameter for memoization
-        :type memo: dict
         :param features: A collection of features or feature types that will be copied into new EOPatch.
-        :type features: object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         """
         if not features:  # For some reason deepcopy and copy pass {} by default
             features = ...
@@ -318,26 +495,22 @@ class EOPatch:
 
         return new_eopatch
 
-    def copy(self, features=..., deep=False):
+    def copy(self, features: FeaturesSpecification = ..., deep: bool = False) -> EOPatch:
         """Get a copy of the current `EOPatch`.
 
         :param features: Features to be copied into a new `EOPatch`. By default, all features will be copied.
-        :type features: object supported by the :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`
         :param deep: If `True` it will make a deep copy of all data inside the `EOPatch`. Otherwise, only a shallow copy
             of `EOPatch` will be made. Note that `BBOX` and `TIMESTAMP` will be copied even with a shallow copy.
-        :type deep: bool
         :return: An EOPatch copy.
-        :rtype: EOPatch
         """
         if deep:
             return self.__deepcopy__(features=features)  # pylint: disable=unnecessary-dunder-call
         return self.__copy__(features=features)  # pylint: disable=unnecessary-dunder-call
 
-    def reset_feature_type(self, feature_type):
+    def reset_feature_type(self, feature_type: FeatureType) -> None:
         """Resets the values of the given feature type.
 
         :param feature_type: Type of feature
-        :type feature_type: FeatureType
         """
         feature_type = FeatureType(feature_type)
         if feature_type.has_dict():
@@ -347,32 +520,29 @@ class EOPatch:
         else:
             self[feature_type] = []
 
-    def get_features(self):
+    def get_features(self) -> Dict[FeatureType, Union[Set[str], Literal[True]]]:
         """Returns a dictionary of all non-empty features of EOPatch.
 
         The elements are either sets of feature names or a boolean `True` in case feature type has no dictionary of
         feature names.
 
         :return: A dictionary of features
-        :rtype: dict(FeatureType: str or True)
         """
-        feature_dict = {}
+        feature_dict: Dict[FeatureType, Union[Set[str], Literal[True]]] = {}
         for feature_type in FeatureType:
             if self[feature_type]:
                 feature_dict[feature_type] = set(self[feature_type]) if feature_type.has_dict() else True
 
         return feature_dict
 
-    def get_spatial_dimension(self, feature_type, feature_name):
+    def get_spatial_dimension(self, feature_type: FeatureType, feature_name: str) -> Tuple[int, int]:
         """
         Returns a tuple of spatial dimension (height, width) of a feature.
 
         The feature has to be spatial or time dependent.
 
-        :param feature_type: Enum of the attribute
-        :type feature_type: FeatureType
+        :param feature_type: Type of the feature
         :param feature_name: Name of the feature
-        :type feature_name: str
         """
         if feature_type.is_temporal() or feature_type.is_spatial():
             shape = self[feature_type][feature_name].shape
@@ -399,23 +569,23 @@ class EOPatch:
         return feature_list
 
     def save(
-        self, path, features=..., overwrite_permission=OverwritePermission.ADD_ONLY, compress_level=0, filesystem=None
-    ):
+        self,
+        path: str,
+        features: FeaturesSpecification = ...,
+        overwrite_permission: OverwritePermission = OverwritePermission.ADD_ONLY,
+        compress_level: int = 0,
+        filesystem: Optional[FS] = None,
+    ) -> None:
         """Method to save an EOPatch from memory to a storage.
 
         :param path: A location where to save EOPatch. It can be either a local path or a remote URL path.
-        :type path: str
         :param features: A collection of features types specifying features of which type will be saved. By default,
             all features will be saved.
-        :type features: list(FeatureType) or list((FeatureType, str)) or ...
         :param overwrite_permission: A level of permission for overwriting an existing EOPatch
-        :type overwrite_permission: OverwritePermission or int
         :param compress_level: A level of data compression and can be specified with an integer from 0 (no compression)
             to 9 (highest compression).
-        :type compress_level: int
         :param filesystem: An existing filesystem object. If not given it will be initialized according to the `path`
             parameter.
-        :type filesystem: fs.FS or None
         """
         if filesystem is None:
             filesystem = get_filesystem(path, create=True)
@@ -431,20 +601,17 @@ class EOPatch:
         )
 
     @staticmethod
-    def load(path, features=..., lazy_loading=False, filesystem=None):
+    def load(
+        path: str, features: FeaturesSpecification = ..., lazy_loading: bool = False, filesystem: Optional[FS] = None
+    ) -> EOPatch:
         """Method to load an EOPatch from a storage into memory.
 
         :param path: A location from where to load EOPatch. It can be either a local path or a remote URL path.
-        :type path: str
         :param features: A collection of features to be loaded. By default, all features will be loaded.
-        :type features: object
         :param lazy_loading: If `True` features will be lazy loaded.
-        :type lazy_loading: bool
         :param filesystem: An existing filesystem object. If not given it will be initialized according to the `path`
             parameter.
-        :type filesystem: fs.FS or None
         :return: Loaded EOPatch
-        :rtype: EOPatch
         """
         if filesystem is None:
             filesystem = get_filesystem(path, create=False)
@@ -452,13 +619,17 @@ class EOPatch:
 
         return load_eopatch(EOPatch(), filesystem, path, features=features, lazy_loading=lazy_loading)
 
-    def merge(self, *eopatches, features=..., time_dependent_op=None, timeless_op=None):
+    def merge(
+        self,
+        *eopatches: EOPatch,
+        features: FeaturesSpecification = ...,
+        time_dependent_op: Union[Literal[None, "concatenate", "min", "max", "mean", "median"], Callable] = None,
+        timeless_op: Union[Literal[None, "concatenate", "min", "max", "mean", "median"], Callable] = None,
+    ) -> EOPatch:
         """Merge features of given EOPatches into a new EOPatch.
 
         :param eopatches: Any number of EOPatches to be merged together with the current EOPatch
-        :type eopatches: EOPatch
         :param features: A collection of features to be merged together. By default, all features will be merged.
-        :type features: object
         :param time_dependent_op: An operation to be used to join data for any time-dependent raster feature. Before
             joining time slices of all arrays will be sorted. Supported options are:
 
@@ -469,7 +640,6 @@ class EOPatch:
             - 'max': Join time slices with matching timestamps by taking maximum values. Ignore NaN values.
             - 'mean': Join time slices with matching timestamps by taking mean values. Ignore NaN values.
             - 'median': Join time slices with matching timestamps by taking median values. Ignore NaN values.
-        :type time_dependent_op: str or Callable or None
         :param timeless_op: An operation to be used to join data for any timeless raster feature. Supported options
             are:
 
@@ -479,9 +649,7 @@ class EOPatch:
             - 'max': Join arrays by taking maximum values. Ignore NaN values.
             - 'mean': Join arrays by taking mean values. Ignore NaN values.
             - 'median': Join arrays by taking median values. Ignore NaN values.
-        :type timeless_op: str or Callable or None
-        :return: A dictionary with EOPatch features and values
-        :rtype: Dict[(FeatureType, str), object]
+        :return: A merged EOPatch
         """
         eopatch_content = merge_eopatches(
             self, *eopatches, features=features, time_dependent_op=time_dependent_op, timeless_op=timeless_op
@@ -493,7 +661,7 @@ class EOPatch:
 
         return merged_eopatch
 
-    def get_time_series(self, ref_date=None, scale_time=1):
+    def get_time_series(self, ref_date: Optional[dt.datetime] = None, scale_time: int = 1) -> np.ndarray:
         """Returns a numpy array with seconds passed between the reference date and the timestamp of each image.
 
         An array is constructed as time_series[i] = (timestamp[i] - ref_date).total_seconds().
@@ -501,9 +669,7 @@ class EOPatch:
         If EOPatch timestamp attribute is empty the method returns None.
 
         :param ref_date: reference date relative to which the time is measured
-        :type ref_date: datetime object
         :param scale_time: scale seconds by factor. If `60`, time will be in minutes, if `3600` hours
-        :type scale_time: int
         """
 
         if not self.timestamp:
@@ -516,13 +682,11 @@ class EOPatch:
             [round((timestamp - ref_date).total_seconds() / scale_time) for timestamp in self.timestamp], dtype=np.int64
         )
 
-    def consolidate_timestamps(self, timestamps):
+    def consolidate_timestamps(self, timestamps: List[dt.datetime]) -> Set[dt.datetime]:
         """Removes all frames from the EOPatch with a date not found in the provided timestamps list.
 
         :param timestamps: keep frames with date found in this list
-        :type timestamps: list of datetime objects
         :return: set of removed frames' dates
-        :rtype: set of datetime objects
         """
         remove_from_patch = set(self.timestamp).difference(timestamps)
         remove_from_patch_idxs = [self.timestamp.index(rm_date) for rm_date in remove_from_patch]
@@ -543,7 +707,7 @@ class EOPatch:
 
     def plot(
         self,
-        feature,
+        feature: FeatureSpec,
         *,
         times: Union[List[int], slice, None] = None,
         channels: Union[List[int], slice, None] = None,
@@ -551,7 +715,7 @@ class EOPatch:
         rgb: Optional[Tuple[int, int, int]] = None,
         backend: Union[str, PlotBackend] = "matplotlib",
         config: Optional[BasePlotConfig] = None,
-        **kwargs,
+        **kwargs: Any,
     ) -> object:
         """Plots an `EOPatch` feature.
 
@@ -587,119 +751,3 @@ class EOPatch:
             config=config,
             **kwargs,
         )
-
-
-class _FeatureDict(dict):
-    """A dictionary structure that holds features of certain feature type.
-
-    It checks that features have a correct and dimension. It also supports lazy loading by accepting a function as a
-    feature value, which is then called when the feature is accessed.
-    """
-
-    FORBIDDEN_CHARS = {".", "/", "\\", "|", ";", ":", "\n", "\t"}
-
-    def __init__(self, feature_dict, feature_type):
-        """
-        :param feature_dict: A dictionary of feature names and values
-        :type feature_dict: dict(str: object)
-        :param feature_type: Type of features
-        :type feature_type: FeatureType
-        """
-        super().__init__()
-
-        self.feature_type = feature_type
-        self.ndim = self.feature_type.ndim()
-        self.is_vector = self.feature_type.is_vector()
-
-        for feature_name, value in feature_dict.items():
-            self[feature_name] = value
-
-    def __setitem__(self, feature_name, value):
-        """Before setting value to the dictionary it checks that value is of correct type and dimension and tries to
-        transform value in correct form.
-        """
-        value = self._parse_feature_value(value, feature_name)
-        self._check_feature_name(feature_name)
-        super().__setitem__(feature_name, value)
-
-    def _check_feature_name(self, feature_name):
-        if not isinstance(feature_name, str):
-            raise ValueError(f"Feature name must be a string but an object of type {type(feature_name)} was given.")
-
-        for char in feature_name:
-            if char in self.FORBIDDEN_CHARS:
-                raise ValueError(
-                    f"The name of feature ({self.feature_type}, {feature_name}) contains an illegal character '{char}'."
-                )
-
-        if feature_name == "":
-            raise ValueError("Feature name cannot be an empty string.")
-
-    def __getitem__(self, feature_name, load=True):
-        """Implements lazy loading."""
-        value = super().__getitem__(feature_name)
-
-        if isinstance(value, FeatureIO) and load:
-            value = value.load()
-            self[feature_name] = value
-
-        return value
-
-    def __eq__(self, other):
-        """Compares its content against a content of another feature type dictionary."""
-        return deep_eq(self, other)
-
-    def __ne__(self, other):
-        """Compares its content against a content of another feature type dictionary."""
-        return not self.__eq__(other)
-
-    def get_dict(self):
-        """Returns a Python dictionary of features and value."""
-        return dict(self)
-
-    def _parse_feature_value(self, value, feature_name):
-        """Checks if value fits the feature type. If not it tries to fix it or raise an error.
-
-        :raises: ValueError
-        """
-        if isinstance(value, FeatureIO):
-            return value
-        if not hasattr(self, "ndim"):  # Because of serialization/deserialization during multiprocessing
-            return value
-
-        if self.ndim:
-            if not isinstance(value, np.ndarray):
-                raise ValueError(f"{self.feature_type} feature has to be a numpy array.")
-            if value.ndim != self.ndim:
-                raise ValueError(
-                    f"Numpy array of {self.feature_type} feature has to have {self.ndim} "
-                    f"dimension{'s' if self.ndim > 1 else ''} but feature {feature_name} has {value.ndim}."
-                )
-
-            if self.feature_type.is_discrete() and not is_discrete_type(value.dtype):
-                raise ValueError(
-                    f"{self.feature_type} is a discrete feature type therefore dtype of data array "
-                    f"has to be either integer or boolean type but feature {feature_name} has dtype {value.dtype.type}."
-                )
-
-            return value
-
-        if self.is_vector:
-            if isinstance(value, gpd.GeoSeries):
-                value = gpd.GeoDataFrame(dict(geometry=value), crs=value.crs)
-
-            if isinstance(value, gpd.GeoDataFrame):
-                if self.feature_type is FeatureType.VECTOR and FeatureType.TIMESTAMP.value.upper() not in value:
-                    raise ValueError(
-                        f"{self.feature_type} feature has to contain a column 'TIMESTAMP' with timestamps but "
-                        f"feature {feature_name} doesn't not have it."
-                    )
-
-                return value
-
-            raise ValueError(
-                f"{self.feature_type} feature works with data of type {gpd.GeoDataFrame.__name__} but feature "
-                f"{feature_name} has data of type {type(value)}."
-            )
-
-        return value
